@@ -4,7 +4,8 @@
  * Copyright (c) 2004-2005 Anton Altaparmakov
  * Copyright (c) 2004-2005 Richard Russon
  * Copyright (c) 2005-2006 Yura Pakhuchiy
- * Copyright (c) 2005-2006 Szabolcs Szakacsits
+ * Copyright (c) 2005-2008 Szabolcs Szakacsits
+ * Copyright (c) 2007 Jean-Pierre Andre
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -37,13 +38,14 @@
 #endif
 
 #include "attrib.h"
-#include "collate.h"
 #include "debug.h"
 #include "index.h"
+#include "collate.h"
 #include "mst.h"
 #include "dir.h"
 #include "logging.h"
 #include "bitmap.h"
+#include "reparse.h"
 #include "misc.h"
 
 /**
@@ -78,18 +80,17 @@ static VCN ntfs_ib_pos_to_vcn(ntfs_index_context *icx, s64 pos)
 	return pos >> icx->vcn_size_bits;
 }
 
-static int ntfs_ib_write(ntfs_index_context *icx, VCN vcn, void *buf)
+static int ntfs_ib_write(ntfs_index_context *icx, INDEX_BLOCK *ib)
 {
-	s64 ret;
+	s64 ret, vcn = sle64_to_cpu(ib->index_block_vcn);
 	
-	ntfs_log_trace("vcn: %lld\n", vcn);
+	ntfs_log_trace("vcn: %lld\n", (long long)vcn);
 	
 	ret = ntfs_attr_mst_pwrite(icx->ia_na, ntfs_ib_vcn_to_pos(icx, vcn),
-				   1, icx->block_size, buf);
+				   1, icx->block_size, ib);
 	if (ret != 1) {
-		ntfs_log_perror("Failed to write index block %lld of inode "
-				"%llu", (long long)vcn,
-				(unsigned long long)icx->ni->mft_no);
+		ntfs_log_perror("Failed to write index block %lld, inode %llu",
+			(long long)vcn, (unsigned long long)icx->ni->mft_no);
 		return STATUS_ERROR;
 	}
 	
@@ -98,7 +99,7 @@ static int ntfs_ib_write(ntfs_index_context *icx, VCN vcn, void *buf)
 
 static int ntfs_icx_ib_write(ntfs_index_context *icx)
 {
-		if (ntfs_ib_write(icx, icx->ib_vcn, icx->ib))
+		if (ntfs_ib_write(icx, icx->ib))
 			return STATUS_ERROR;
 		
 		icx->ib_dirty = FALSE;
@@ -148,18 +149,14 @@ static void ntfs_index_ctx_free(ntfs_index_context *icx)
 	if (icx->actx)
 		ntfs_attr_put_search_ctx(icx->actx);
 
-	if (icx->is_in_root) {
-		if (icx->ia_na)
-			ntfs_attr_close(icx->ia_na);
-		return;
-	}
-
-	if (icx->ib_dirty) {
-		/* FIXME: Error handling!!! */
-		ntfs_ib_write(icx, icx->ib_vcn, icx->ib);
+	if (!icx->is_in_root) {
+		if (icx->ib_dirty) {
+			/* FIXME: Error handling!!! */
+			ntfs_ib_write(icx, icx->ib);
+		}
+		free(icx->ib);
 	}
 	
-	free(icx->ib);
 	ntfs_attr_close(icx->ia_na);
 }
 
@@ -225,7 +222,7 @@ static u8 *ntfs_ie_get_end(INDEX_HEADER *ih)
 
 static int ntfs_ie_end(INDEX_ENTRY *ie)
 {
-	return ie->ie_flags & INDEX_ENTRY_END;
+	return ie->ie_flags & INDEX_ENTRY_END || !ie->length;
 }
 
 /** 
@@ -306,11 +303,13 @@ static int ntfs_ih_numof_entries(INDEX_HEADER *ih)
 {
 	int n;
 	INDEX_ENTRY *ie;
+	u8 *end;
 	
 	ntfs_log_trace("Entering\n");
 	
+	end = ntfs_ie_get_end(ih);
 	ie = ntfs_ie_get_first(ih);
-	for (n = 0; !ntfs_ie_end(ie); n++)
+	for (n = 0; !ntfs_ie_end(ie) && (u8 *)ie < end; n++)
 		ie = ntfs_ie_get_next(ie);
 	return n;
 }
@@ -434,10 +433,8 @@ static INDEX_ROOT *ntfs_ir_lookup(ntfs_inode *ni, ntfschar *name,
 	ntfs_log_trace("Entering\n");
 	
 	*ctx = ntfs_attr_get_search_ctx(ni, NULL);
-	if (!*ctx) {
-		ntfs_log_perror("Failed to get $INDEX_ROOT search context");
+	if (!*ctx)
 		return NULL;
-	}
 	
 	if (ntfs_attr_lookup(AT_INDEX_ROOT, name, name_len, CASE_SENSITIVE, 
 			     0, NULL, 0, *ctx)) {
@@ -454,8 +451,21 @@ static INDEX_ROOT *ntfs_ir_lookup(ntfs_inode *ni, ntfschar *name,
 	
 	ir = (INDEX_ROOT *)((char *)a + le16_to_cpu(a->value_offset));
 err_out:
-	if (!ir)
+	if (!ir) {
 		ntfs_attr_put_search_ctx(*ctx);
+		*ctx = NULL;
+	}
+	return ir;
+}
+
+static INDEX_ROOT *ntfs_ir_lookup2(ntfs_inode *ni, ntfschar *name, u32 len)
+{
+	ntfs_attr_search_ctx *ctx;
+	INDEX_ROOT *ir;
+
+	ir = ntfs_ir_lookup(ni, name, len, &ctx);
+	if (ir)
+		ntfs_attr_put_search_ctx(ctx);
 	return ir;
 }
 
@@ -507,8 +517,13 @@ static int ntfs_ie_lookup(const void *key, const int key_len,
 		 * Not a perfect match, need to do full blown collation so we
 		 * know which way in the B+tree we have to go.
 		 */
-		rc = ntfs_collate(icx->ni->vol, icx->cr, key, key_len, &ie->key,
-				  le16_to_cpu(ie->key_length));
+		if (!icx->collate) {
+			ntfs_log_error("Collation function not defined\n");
+			errno = EOPNOTSUPP;
+			return STATUS_ERROR;
+		}
+		rc = icx->collate(icx->ni->vol, key, key_len,
+					&ie->key, le16_to_cpu(ie->key_length));
 		if (rc == NTFS_COLLATION_ERROR) {
 			ntfs_log_error("Collation error. Perhaps a filename "
 				       "contains invalid characters?\n");
@@ -548,7 +563,7 @@ static int ntfs_ie_lookup(const void *key, const int key_len,
 	*vcn = ntfs_ie_get_vcn(ie);
 	if (*vcn < 0) {
 		errno = EINVAL;
-		ntfs_log_perror("Negative vcn in inode %llu\n",
+		ntfs_log_perror("Negative vcn in inode %llu",
 			       	(unsigned long long)icx->ni->mft_no);
 		return STATUS_ERROR;
 	}
@@ -577,7 +592,7 @@ static int ntfs_ib_read(ntfs_index_context *icx, VCN vcn, INDEX_BLOCK *dst)
 {
 	s64 pos, ret;
 
-	ntfs_log_trace("vcn: %lld\n", vcn);
+	ntfs_log_trace("vcn: %lld\n", (long long)vcn);
 	
 	pos = ntfs_ib_vcn_to_pos(icx, vcn);
 
@@ -651,15 +666,13 @@ static int ntfs_icx_parent_dec(ntfs_index_context *icx)
  * the call to ntfs_index_ctx_put() to ensure that the changes are written
  * to disk.
  */
-int ntfs_index_lookup(const void *key, const int key_len,
-		ntfs_index_context *icx)
+int ntfs_index_lookup(const void *key, const int key_len, ntfs_index_context *icx)
 {
 	VCN old_vcn, vcn;
 	ntfs_inode *ni = icx->ni;
 	INDEX_ROOT *ir;
 	INDEX_ENTRY *ie;
 	INDEX_BLOCK *ib = NULL;
-	ntfs_attr_search_ctx *actx;
 	int ret, err = 0;
 
 	ntfs_log_trace("Entering\n");
@@ -670,7 +683,7 @@ int ntfs_index_lookup(const void *key, const int key_len,
 		return -1;
 	}
 
-	ir = ntfs_ir_lookup(ni, icx->name, icx->name_len, &actx);
+	ir = ntfs_ir_lookup(ni, icx->name, icx->name_len, &icx->actx);
 	if (!ir) {
 		if (errno == ENOENT)
 			errno = EIO;
@@ -682,19 +695,19 @@ int ntfs_index_lookup(const void *key, const int key_len,
 		errno = EINVAL;
 		ntfs_log_perror("Index block size (%d) is smaller than the "
 				"sector size (%d)", icx->block_size, NTFS_BLOCK_SIZE);
-		return -1;
+		goto err_out;
 	}
 
 	if (ni->vol->cluster_size <= icx->block_size)
 		icx->vcn_size_bits = ni->vol->cluster_size_bits;
 	else
 		icx->vcn_size_bits = ni->vol->sector_size_bits;
-		
-	icx->cr = ir->collation_rule;
-	if (!ntfs_is_collation_rule_supported(icx->cr)) {
+			/* get the appropriate collation function */
+	icx->collate = ntfs_get_collate_function(ir->collation_rule);
+	if (!icx->collate) {
 		err = errno = EOPNOTSUPP;
 		ntfs_log_perror("Unknown collation rule 0x%x", 
-				(unsigned)le32_to_cpu(icx->cr));
+				(unsigned)le32_to_cpu(ir->collation_rule));
 		goto err_out;
 	}
 	
@@ -709,7 +722,6 @@ int ntfs_index_lookup(const void *key, const int key_len,
 		goto err_out;
 	}
 	
-	icx->actx = actx;
 	icx->ir = ir;
 	
 	if (ret != STATUS_KEEP_SEARCHING) {
@@ -741,7 +753,7 @@ descend_into_child_node:
 	}
 	old_vcn = vcn;
 
-	ntfs_log_debug("Descend into node with VCN %lld.\n", vcn);
+	ntfs_log_debug("Descend into node with VCN %lld\n", (long long)vcn);
 	
 	if (ntfs_ib_read(icx, vcn, ib))
 		goto err_out;
@@ -755,7 +767,7 @@ descend_into_child_node:
 		/* STATUS_OK or STATUS_NOT_FOUND */
 		icx->is_in_root = FALSE;
 		icx->ib = ib;
-		icx->parent_vcn[icx->pindex] = icx->ib_vcn = vcn;
+		icx->parent_vcn[icx->pindex] = vcn;
 		goto done;
 	}
 
@@ -768,20 +780,15 @@ descend_into_child_node:
 	
 	goto descend_into_child_node;
 err_out:
-	if (icx->ia_na)
-		ntfs_attr_close(icx->ia_na);
 	free(ib);
 	if (!err)
 		err = EIO;
-	if (actx)
-		ntfs_attr_put_search_ctx(actx);
 	errno = err;
 	return -1;
 done:
 	icx->entry = ie;
 	icx->data = (u8 *)ie + offsetof(INDEX_ENTRY, key);
 	icx->data_len = le16_to_cpu(ie->key_length);
-	icx->max_depth = icx->pindex;
 	ntfs_log_trace("Done.\n");
 	if (err) {
 		errno = err;
@@ -797,7 +804,7 @@ static INDEX_BLOCK *ntfs_ib_alloc(VCN ib_vcn, u32 ib_size,
 	INDEX_BLOCK *ib;
 	int ih_size = sizeof(INDEX_HEADER);
 	
-	ntfs_log_trace("Entering ib_vcn = %lld ib_size = %u\n", ib_vcn, ib_size);
+	ntfs_log_trace("ib_vcn: %lld ib_size: %u\n", (long long)ib_vcn, ib_size);
 	
 	ib = ntfs_calloc(ib_size);
 	if (!ib)
@@ -893,7 +900,7 @@ static int ntfs_ibm_modify(ntfs_index_context *icx, VCN vcn, int set)
 	ntfs_attr *na;
 	int ret = STATUS_ERROR;
 
-	ntfs_log_trace("%s vcn: %lld\n", set ? "set" : "clear", vcn);
+	ntfs_log_trace("%s vcn: %lld\n", set ? "set" : "clear", (long long)vcn);
 	
 	na = ntfs_attr_open(icx->ni, AT_BITMAP,  icx->name, icx->name_len);
 	if (!na) {
@@ -970,7 +977,7 @@ static VCN ntfs_ibm_get_free(ntfs_index_context *icx)
 	
 	vcn = ntfs_ibm_pos_to_vcn(icx, size * 8);
 out:	
-	ntfs_log_trace("allocated vcn: %lld\n", vcn);
+	ntfs_log_trace("allocated vcn: %lld\n", (long long)vcn);
 
 	if (ntfs_ibm_set(icx, vcn))
 		vcn = (VCN)-1;
@@ -1052,13 +1059,13 @@ static int ntfs_ib_copy_tail(ntfs_index_context *icx, INDEX_BLOCK *src,
 	
 	dst->index.index_length = cpu_to_le32(tail_size + 
 					      le32_to_cpu(dst->index.entries_offset));
-	ret = ntfs_ib_write(icx, new_vcn, dst);
+	ret = ntfs_ib_write(icx, dst);
 
 	free(dst);
 	return ret;
 }
 
-static int ntfs_ib_cut_tail(ntfs_index_context *icx, INDEX_BLOCK *src,
+static int ntfs_ib_cut_tail(ntfs_index_context *icx, INDEX_BLOCK *ib,
 			    INDEX_ENTRY *ie)
 {
 	char *ies_start, *ies_end;
@@ -1066,8 +1073,8 @@ static int ntfs_ib_cut_tail(ntfs_index_context *icx, INDEX_BLOCK *src,
 	
 	ntfs_log_trace("Entering\n");
 	
-	ies_start = (char *)ntfs_ie_get_first(&src->index);
-	ies_end   = (char *)ntfs_ie_get_end(&src->index);
+	ies_start = (char *)ntfs_ie_get_first(&ib->index);
+	ies_end   = (char *)ntfs_ie_get_end(&ib->index);
 	
 	ie_last   = ntfs_ie_get_last((INDEX_ENTRY *)ies_start, ies_end);
 	if (ie_last->ie_flags & INDEX_ENTRY_NODE)
@@ -1075,10 +1082,10 @@ static int ntfs_ib_cut_tail(ntfs_index_context *icx, INDEX_BLOCK *src,
 	
 	memcpy(ie, ie_last, le16_to_cpu(ie_last->length));
 	
-	src->index.index_length = cpu_to_le32(((char *)ie - ies_start) + 
-		le16_to_cpu(ie->length) + le32_to_cpu(src->index.entries_offset));
+	ib->index.index_length = cpu_to_le32(((char *)ie - ies_start) + 
+		le16_to_cpu(ie->length) + le32_to_cpu(ib->index.entries_offset));
 	
-	if (ntfs_ib_write(icx, icx->parent_vcn[icx->pindex + 1], src))
+	if (ntfs_ib_write(icx, ib))
 		return STATUS_ERROR;
 	
 	return STATUS_OK;
@@ -1109,7 +1116,7 @@ static int ntfs_ia_add(ntfs_index_context *icx)
 
 static int ntfs_ir_reparent(ntfs_index_context *icx)
 {
-	ntfs_attr_search_ctx *ctx;
+	ntfs_attr_search_ctx *ctx = NULL;
 	INDEX_ROOT *ir;
 	INDEX_ENTRY *ie;
 	INDEX_BLOCK *ib = NULL;
@@ -1118,25 +1125,33 @@ static int ntfs_ir_reparent(ntfs_index_context *icx)
 
 	ntfs_log_trace("Entering\n");
 	
-	ir = ntfs_ir_lookup(icx->ni, icx->name, icx->name_len, &ctx);
+	ir = ntfs_ir_lookup2(icx->ni, icx->name, icx->name_len);
 	if (!ir)
-		return -1;
+		goto out;
 	
 	if ((ir->index.ih_flags & NODE_MASK) == SMALL_INDEX)
 		if (ntfs_ia_add(icx))
-			goto err_out;
+			goto out;
 	
 	new_ib_vcn = ntfs_ibm_get_free(icx);
 	if (new_ib_vcn == -1)
-		goto err_out;
+		goto out;
 		
+	ir = ntfs_ir_lookup2(icx->ni, icx->name, icx->name_len);
+	if (!ir)
+		goto clear_bmp;
+	
 	ib = ntfs_ir_to_ib(ir, new_ib_vcn);
 	if (ib == NULL) {
 		ntfs_log_perror("Failed to move index root to index block");
 		goto clear_bmp;
 	}
 		
-	if (ntfs_ib_write(icx, new_ib_vcn, ib))
+	if (ntfs_ib_write(icx, ib))
+		goto clear_bmp;
+	
+	ir = ntfs_ir_lookup(icx->ni, icx->name, icx->name_len, &ctx);
+	if (!ir)
 		goto clear_bmp;
 	
 	ntfs_ir_nill(ir);
@@ -1153,8 +1168,8 @@ static int ntfs_ir_reparent(ntfs_index_context *icx)
 	if (ntfs_resident_attr_value_resize(ctx->mrec, ctx->attr,
 			sizeof(INDEX_ROOT) - sizeof(INDEX_HEADER) +
 			le32_to_cpu(ir->index.allocated_size)))
-		/* FIXME: revert bitmap, index root */
-		goto err_out;
+		/* FIXME: revert index root */
+		goto clear_bmp;
 	/*
 	 *  FIXME: do it earlier if we have enough space in IR (should always),
 	 *  so in error case we wouldn't lose the IB.
@@ -1165,6 +1180,7 @@ static int ntfs_ir_reparent(ntfs_index_context *icx)
 err_out:
 	free(ib);
 	ntfs_attr_put_search_ctx(ctx);
+out:
 	return ret;
 clear_bmp:
 	ntfs_ibm_clear(icx, new_ib_vcn);
@@ -1195,18 +1211,14 @@ static int ntfs_ir_truncate(ntfs_index_context *icx, int data_size)
 	ret = ntfs_attr_truncate(na, data_size + offsetof(INDEX_ROOT, index));
 	if (ret == STATUS_OK) {
 		
-		ntfs_attr_search_ctx *ctx;
-		
-		icx->ir = ntfs_ir_lookup(icx->ni, icx->name, icx->name_len, &ctx);
+		icx->ir = ntfs_ir_lookup2(icx->ni, icx->name, icx->name_len);
 		if (!icx->ir)
 			return STATUS_ERROR;
 	
 		icx->ir->index.allocated_size = cpu_to_le32(data_size);
 		
-		ntfs_attr_put_search_ctx(ctx);
-	
-	} else if (errno != ENOSPC)
-		ntfs_log_trace("Failed to truncate INDEX_ROOT");
+	} else if (ret == STATUS_ERROR)
+		ntfs_log_perror("Failed to truncate INDEX_ROOT");
 	
 	ntfs_attr_close(na);
 	return ret;
@@ -1223,16 +1235,16 @@ static int ntfs_ir_make_space(ntfs_index_context *icx, int data_size)
 	int ret;
 
 	ntfs_log_trace("Entering\n");
-	
+
 	ret = ntfs_ir_truncate(icx, data_size);
 	if (ret == STATUS_RESIDENT_ATTRIBUTE_FILLED_MFT) {
-		
+	
 		ret = ntfs_ir_reparent(icx);
 		if (ret == STATUS_OK)
 			ret = STATUS_KEEP_SEARCHING;
 		else
 			ntfs_log_perror("Failed to nodify INDEX_ROOT");
-	} 
+	}
 
 	return ret;
 }
@@ -1304,6 +1316,10 @@ static int ntfs_ir_insert_median(ntfs_index_context *icx, INDEX_ENTRY *median,
 	
 	ntfs_log_trace("Entering\n");
 	
+	icx->ir = ntfs_ir_lookup2(icx->ni, icx->name, icx->name_len);
+	if (!icx->ir)
+		return STATUS_ERROR;
+
 	new_size = le32_to_cpu(icx->ir->index.index_length) + 
 			le16_to_cpu(median->length);
 	if (!(median->ie_flags & INDEX_ENTRY_NODE))
@@ -1312,6 +1328,10 @@ static int ntfs_ir_insert_median(ntfs_index_context *icx, INDEX_ENTRY *median,
 	ret = ntfs_ir_make_space(icx, new_size);
 	if (ret != STATUS_OK)
 		return ret;
+	
+	icx->ir = ntfs_ir_lookup2(icx->ni, icx->name, icx->name_len);
+	if (!icx->ir)
+		return STATUS_ERROR;
 
 	return ntfs_ih_insert(&icx->ir->index, median, new_vcn, 
 			      ntfs_icx_parent_pos(icx));
@@ -1354,7 +1374,7 @@ static int ntfs_ib_insert(ntfs_index_context *icx, INDEX_ENTRY *ie, VCN new_vcn)
 	if (ntfs_ih_insert(&ib->index, ie, new_vcn, ntfs_icx_parent_pos(icx)))
 		goto err_out;
 	
-	if (ntfs_ib_write(icx, old_vcn, ib))
+	if (ntfs_ib_write(icx, ib))
 		goto err_out;
 	
 	err = STATUS_OK;
@@ -1364,7 +1384,7 @@ err_out:
 }
 
 /**
- * ntfs_ib_split - Split index allocation attribute
+ * ntfs_ib_split - Split an index block
  * 
  * On success return STATUS_OK or STATUS_KEEP_SEARCHING.
  * On error return is STATUS_ERROR.
@@ -1395,8 +1415,6 @@ static int ntfs_ib_split(ntfs_index_context *icx, INDEX_BLOCK *ib)
 	else
 		ret = ntfs_ib_insert(icx, median, new_vcn);
 	
-	ntfs_inode_mark_dirty(icx->actx->ntfs_ino);
-	
 	if (ret != STATUS_OK) {
 		ntfs_ibm_clear(icx, new_vcn);
 		return ret;
@@ -1407,16 +1425,21 @@ static int ntfs_ib_split(ntfs_index_context *icx, INDEX_BLOCK *ib)
 	return ret;
 }
 
-
-static int ntfs_ie_add(ntfs_index_context *icx, INDEX_ENTRY *ie)
+/* JPA static */
+int ntfs_ie_add(ntfs_index_context *icx, INDEX_ENTRY *ie)
 {
-	char *fn;
 	INDEX_HEADER *ih;
 	int allocated_size, new_size;
 	int ret = STATUS_ERROR;
 	
+#ifdef DEBUG
+/* removed by JPA to make function usable for security indexes
+	char *fn;
 	fn = ntfs_ie_filename_get(ie);
 	ntfs_log_trace("file: '%s'\n", fn);
+	ntfs_attr_name_free(&fn);
+*/
+#endif
 	
 	while (1) {
 				
@@ -1461,7 +1484,6 @@ static int ntfs_ie_add(ntfs_index_context *icx, INDEX_ENTRY *ie)
 	
 	ret = STATUS_OK;
 err_out:
-	ntfs_attr_name_free(&fn);
 	ntfs_log_trace("%s\n", ret ? "Failed" : "Done");
 	return ret;
 }
@@ -1478,7 +1500,7 @@ int ntfs_index_add_filename(ntfs_inode *ni, FILE_NAME_ATTR *fn, MFT_REF mref)
 {
 	INDEX_ENTRY *ie;
 	ntfs_index_context *icx;
-	int fn_size, ie_size, ret = -1;
+	int fn_size, ie_size, err, ret = -1;
 
 	ntfs_log_trace("Entering\n");
 	
@@ -1506,8 +1528,9 @@ int ntfs_index_add_filename(ntfs_inode *ni, FILE_NAME_ATTR *fn, MFT_REF mref)
 		goto out;
 	
 	ret = ntfs_ie_add(icx, ie);
-	
+	err = errno;
 	ntfs_index_ctx_put(icx);
+	errno = err;
 out:
 	free(ie);
 	return ret;
@@ -1530,7 +1553,7 @@ static int ntfs_ih_takeout(ntfs_index_context *icx, INDEX_HEADER *ih,
 	if (ntfs_icx_parent_vcn(icx) == VCN_INDEX_ROOT_PARENT)
 		ntfs_inode_mark_dirty(icx->actx->ntfs_ino);
 	else
-		if (ntfs_ib_write(icx, ntfs_icx_parent_vcn(icx), ib))
+		if (ntfs_ib_write(icx, ib))
 			goto out;
 	
 	ntfs_index_ctx_reinit(icx);
@@ -1560,9 +1583,6 @@ static void ntfs_ir_leafify(ntfs_index_context *icx, INDEX_HEADER *ih)
 	
 	/* Not fatal error */
 	ntfs_ir_truncate(icx, le32_to_cpu(ih->index_length));
-	
-	ntfs_inode_mark_dirty(icx->actx->ntfs_ino);
-	ntfs_index_ctx_reinit(icx);
 }
 
 /**
@@ -1640,7 +1660,7 @@ out:
 
 static int ntfs_index_rm_node(ntfs_index_context *icx)
 {
-	int entry_pos;
+	int entry_pos, pindex;
 	VCN vcn;
 	INDEX_BLOCK *ib = NULL;
 	INDEX_ENTRY *ie_succ, *ie, *entry = icx->entry;
@@ -1662,6 +1682,7 @@ static int ntfs_index_rm_node(ntfs_index_context *icx)
 	
 	ie_succ = ntfs_ie_get_next(icx->entry);
 	entry_pos = icx->parent_pos[icx->pindex]++;
+	pindex = icx->pindex;
 descend:
 	vcn = ntfs_ie_get_vcn(ie_succ);
 	if (ntfs_ib_read(icx, vcn, ib))
@@ -1679,9 +1700,8 @@ descend:
 		goto descend;
 
 	if (ntfs_ih_zero_entry(&ib->index)) {
-		errno = EOPNOTSUPP;
-		ntfs_log_perror("Failed to find any entry in an index block. "
-				"Please run chkdsk.");
+		errno = EIO;
+		ntfs_log_perror("Empty index block");
 		goto out;
 	}
 
@@ -1703,18 +1723,18 @@ descend:
 	new_size = le32_to_cpu(ih->index_length) + delta;
 	if (delta > 0) {
 		if (icx->is_in_root) {
-			if (ntfs_ir_truncate(icx, new_size)) {
-				errno = EOPNOTSUPP;
-				ntfs_log_perror("Denied to truncate INDEX ROOT during entry removal");
+			ret = ntfs_ir_make_space(icx, new_size);
+			if (ret != STATUS_OK)
 				goto out2;
-			}
 			
 			ih = &icx->ir->index;
 			entry = ntfs_ie_get_by_pos(ih, entry_pos);
 			
 		} else if (new_size > le32_to_cpu(ih->allocated_size)) {
-			errno = EOPNOTSUPP;
-			ntfs_log_perror("Denied to split INDEX BLOCK during entry removal");
+			icx->pindex = pindex;
+			ret = ntfs_ib_split(icx, icx->ib);
+			if (ret == STATUS_OK)
+				ret = STATUS_KEEP_SEARCHING;
 			goto out2;
 		}
 	}
@@ -1725,7 +1745,6 @@ descend:
 	if (icx->is_in_root) {
 		if (ntfs_ir_truncate(icx, new_size))
 			goto out2;
-		ntfs_inode_mark_dirty(icx->actx->ntfs_ino);
 	} else
 		if (ntfs_icx_ib_write(icx))
 			goto out2;
@@ -1736,7 +1755,7 @@ descend:
 		if (ntfs_index_rm_leaf(icx))
 			goto out2;
 	} else 
-		if (ntfs_ib_write(icx, vcn, ib))
+		if (ntfs_ib_write(icx, ib))
 			goto out2;
 
 	ret = STATUS_OK;
@@ -1757,10 +1776,11 @@ out:
  *
  * Return 0 on success or -1 on error with errno set to the error code.
  */
+/*static JPA*/
 int ntfs_index_rm(ntfs_index_context *icx)
 {
 	INDEX_HEADER *ih;
-	int err;
+	int err, ret = STATUS_OK;
 
 	ntfs_log_trace("Entering\n");
 	
@@ -1776,8 +1796,7 @@ int ntfs_index_rm(ntfs_index_context *icx)
 	
 	if (icx->entry->ie_flags & INDEX_ENTRY_NODE) {
 		
-		if (ntfs_index_rm_node(icx))
-			goto err_out;
+		ret = ntfs_index_rm_node(icx);
 
 	} else if (icx->is_in_root || !ntfs_ih_one_entry(ih)) {
 		
@@ -1794,16 +1813,53 @@ int ntfs_index_rm(ntfs_index_context *icx)
 		if (ntfs_index_rm_leaf(icx))
 			goto err_out;
 	}
-		
-	ntfs_index_ctx_reinit(icx);
-	ntfs_log_trace("Done.\n");
-	return 0;
+out:
+	return ret;
 err_out:
-	err = errno;
-	ntfs_index_ctx_reinit(icx);
-	errno = err;
-	ntfs_log_trace("Failed.\n");
-	return -1;
+	ret = STATUS_ERROR;
+	goto out;
+}
+
+int ntfs_index_remove(ntfs_inode *dir_ni, ntfs_inode *ni,
+		const void *key, const int keylen)
+{
+	int ret = STATUS_ERROR;
+	ntfs_index_context *icx;
+
+	icx = ntfs_index_ctx_get(dir_ni, NTFS_INDEX_I30, 4);
+	if (!icx)
+		return -1;
+
+	while (1) {
+				
+		if (ntfs_index_lookup(key, keylen, icx))
+			goto err_out;
+
+		if ((((FILE_NAME_ATTR *)icx->data)->file_attributes &
+				FILE_ATTR_REPARSE_POINT)
+		   && !ntfs_possible_symlink(ni)) {
+			errno = EOPNOTSUPP;
+			goto err_out;
+		}
+
+		ret = ntfs_index_rm(icx);
+		if (ret == STATUS_ERROR)
+			goto err_out;
+		else if (ret == STATUS_OK)
+			break;
+		
+		ntfs_inode_mark_dirty(icx->actx->ntfs_ino);
+		ntfs_index_ctx_reinit(icx);
+	}
+
+	ntfs_inode_mark_dirty(icx->actx->ntfs_ino);
+out:	
+	ntfs_index_ctx_put(icx);
+	return ret;
+err_out:
+	ret = STATUS_ERROR;
+	ntfs_log_perror("Delete failed");
+	goto out;
 }
 
 /**
@@ -1841,3 +1897,167 @@ out:
 }
 
 
+/*
+ *		Walk down the index tree (leaf bound)
+ *	until there are no subnode in the first index entry
+ *	returns the entry at the bottom left in subnode
+ */
+
+static INDEX_ENTRY *ntfs_index_walk_down(INDEX_ENTRY *ie,
+			ntfs_index_context *ictx)
+{
+	INDEX_ENTRY *entry;
+	s64 vcn;
+
+	entry = ie;
+	do {
+		vcn = ntfs_ie_get_vcn(entry);
+		if (ictx->is_in_root) {
+
+			/* down from level zero */
+
+			ictx->ir = (INDEX_ROOT*)NULL;
+			ictx->ib = (INDEX_BLOCK*)ntfs_malloc(ictx->block_size);
+			ictx->pindex = 1;
+			ictx->is_in_root = FALSE;
+		} else {
+
+			/* down from non-zero level */
+			
+			ictx->pindex++;
+		}
+		ictx->parent_pos[ictx->pindex] = 0;
+		ictx->parent_vcn[ictx->pindex] = vcn;
+		if (!ntfs_ib_read(ictx,vcn,ictx->ib)) {
+			ictx->entry = ntfs_ie_get_first(&ictx->ib->index);
+			entry = ictx->entry;
+		} else
+			entry = (INDEX_ENTRY*)NULL;
+	} while (entry && (entry->ie_flags & INDEX_ENTRY_NODE));
+	return (entry);
+}
+
+/*
+ *		Walk up the index tree (root bound)
+ *	until there is a valid data entry in parent
+ *	returns the parent entry or NULL if no more parent
+ */
+
+static INDEX_ENTRY *ntfs_index_walk_up(INDEX_ENTRY *ie,
+			ntfs_index_context *ictx)
+{
+	INDEX_ENTRY *entry;
+	s64 vcn;
+
+	entry = ie;
+	if (ictx->pindex > 0) {
+		do {
+			ictx->pindex--;
+			if (!ictx->pindex) {
+
+					/* we have reached the root */
+
+				free(ictx->ib);
+				ictx->ib = (INDEX_BLOCK*)NULL;
+				ictx->is_in_root = TRUE;
+				/* a new search context is to be allocated */
+				if (ictx->actx)
+					free(ictx->actx);
+				ictx->ir = ntfs_ir_lookup(ictx->ni,
+					ictx->name, ictx->name_len,
+					&ictx->actx);
+				if (ictx->ir)
+					entry = ntfs_ie_get_by_pos(
+						&ictx->ir->index,
+						ictx->parent_pos[ictx->pindex]);
+				else
+					entry = (INDEX_ENTRY*)NULL;
+			} else {
+					/* up into non-root node */
+				vcn = ictx->parent_vcn[ictx->pindex];
+				if (!ntfs_ib_read(ictx,vcn,ictx->ib)) {
+					entry = ntfs_ie_get_by_pos(
+						&ictx->ib->index,
+						ictx->parent_pos[ictx->pindex]);
+				} else
+					entry = (INDEX_ENTRY*)NULL;
+			}
+		ictx->entry = entry;
+		} while (entry && (ictx->pindex > 0)
+			 && (entry->ie_flags & INDEX_ENTRY_END));
+	} else
+		entry = (INDEX_ENTRY*)NULL;
+	return (entry);
+}
+
+/*
+ *		Get next entry in an index according to collating sequence.
+ *	Must be initialized through a ntfs_index_lookup()
+ *
+ *	Returns next entry or NULL if none
+ *
+ *	Sample layout :
+ *
+ *                 +---+---+---+---+---+---+---+---+    n ptrs to subnodes
+ *                 |   |   | 10| 25| 33|   |   |   |    n-1 keys in between
+ *                 +---+---+---+---+---+---+---+---+    no key in last entry
+ *                              | A | A
+ *                              | | | +-------------------------------+
+ *   +--------------------------+ | +-----+                           |
+ *   |                            +--+    |                           |
+ *   V                               |    V                           |
+ * +---+---+---+---+---+---+---+---+ |  +---+---+---+---+---+---+---+---+
+ * | 11| 12| 13| 14| 15| 16| 17|   | |  | 26| 27| 28| 29| 30| 31| 32|   |
+ * +---+---+---+---+---+---+---+---+ |  +---+---+---+---+---+---+---+---+
+ *                               |   |
+ *       +-----------------------+   |
+ *       |                           |
+ *     +---+---+---+---+---+---+---+---+
+ *     | 18| 19| 20| 21| 22| 23| 24|   |
+ *     +---+---+---+---+---+---+---+---+
+ */
+
+INDEX_ENTRY *ntfs_index_next(INDEX_ENTRY *ie, ntfs_index_context *ictx)
+{
+	INDEX_ENTRY *next;
+	int flags;
+
+			/*
+                         * lookup() may have returned an invalid node
+			 * when searching for a partial key
+			 * if this happens, walk up
+			 */
+
+	if (ie->ie_flags & INDEX_ENTRY_END)
+		next = ntfs_index_walk_up(ie, ictx);
+	else {
+			/*
+			 * get next entry in same node
+                         * there is always one after any entry with data
+			 */
+
+		next = (INDEX_ENTRY*)((char*)ie + le16_to_cpu(ie->length));
+		++ictx->parent_pos[ictx->pindex];
+		flags = next->ie_flags;
+
+			/* walk down if it has a subnode */
+
+		if (flags & INDEX_ENTRY_NODE) {
+			next = ntfs_index_walk_down(next,ictx);
+		} else {
+
+				/* walk up it has no subnode, nor data */
+
+			if (flags & INDEX_ENTRY_END) {
+				next = ntfs_index_walk_up(next, ictx);
+			}
+		}
+	}
+		/* return NULL if stuck at end of a block */
+
+	if (next && (next->ie_flags & INDEX_ENTRY_END))
+		next = (INDEX_ENTRY*)NULL;
+	return (next);
+}
+
+
